#include "mipsample.h"
#include "global.h"
#include "GL/glew.h"
#include "SDL.h"
#include "timer.h"
#include "color_space.h"

#include "uniform_arrays.h"
#include "screengrabber.h"

#include <fstream>

#define USE_SRGB

using namespace std;

//int screen_width = 89;
//int screen_height = 89;
int screen_width = 512;
int screen_height = 512;
int filter_mode = 0;
Array<string> filter_names;
Array<string> texture_names;

MovieGrabber movie;


int iters = 1;

template<class MMM, class FFF>
void resample(FFF &filter, MMM &mip, Array2D<vect4f> &dst, vect2f &trans, vect<2, vect2f> &mat)
{
	float scale = mip.s[0] / (double)dst.s[0];
	float w = log(scale) / log(2.0);

	for (int i = 0; i < dst.s[0]; i++)
	{
		for (int j = 0; j < dst.s[1]; j++)
		{
			vect2f p, q;
			p.set((i + .5) / dst.s[0], (j + .5) / dst.s[1]);
			p -= .5;
			p -= trans;
			q[0] = mat[0]*p;
			q[1] = mat[1]*p;
			q += .5;

			dst(i,j) = filter(mip, q[0], q[1], w);
			
			for (int k = 0; k < 4; k++)
				dst(i,j)[k] = max(0.0f,min(1.0f,dst(i,j)[k]));
		}
	}
}

void load_mesh(string fn)
{
	// read from file
	g.mesh.load(fn);
	//g.mesh.scaleToBox();

	printf("loaded mesh %s with %d tris\n", fn.c_str(), g.mesh.tris.s);

	// convert data
	Array<VertexData> verts;
	verts.resize(g.mesh.tris.s * 3);

	for (int i = 0; i < g.mesh.tris.s; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			verts[i*3 + j].p_x = g.mesh.pos[g.mesh.tris[i].p[j]][0];
			verts[i*3 + j].p_y = g.mesh.pos[g.mesh.tris[i].p[j]][1];
			verts[i*3 + j].p_z = g.mesh.pos[g.mesh.tris[i].p[j]][2];
			
			verts[i*3 + j].c_u = g.mesh.uv[g.mesh.tris[i].t[j]][0];
			verts[i*3 + j].c_v = g.mesh.uv[g.mesh.tris[i].t[j]][1];
		}
	}
	
	// upload it
	g.mesh_vb_size = verts.s;

	glGenBuffers(1, &g.mesh_vb);
	glBindBuffer(GL_ARRAY_BUFFER, g.mesh_vb);
	glBufferData(GL_ARRAY_BUFFER, g.mesh_vb_size * sizeof(VertexData), verts.data, GL_STATIC_DRAW);

	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void*)0);
	glEnableVertexAttribArray(0);

	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void*)12);
	glEnableVertexAttribArray(1);
}

void create_quad()
{
	// create vertex buffer
	VertexData verts[6];

	int r = 100;

	verts[0].p_x = -1 - 2*(r-1);
	verts[0].p_y = -1 - 2*(r-1);
	verts[0].p_z = 0;
	verts[0].c_u = -(r-1);
	verts[0].c_v = -(r-1);

	verts[1].p_x = 2*r - 1;
	verts[1].p_y = -1 - 2*(r-1);
	verts[1].p_z = 0;
	verts[1].c_u = r;
	verts[1].c_v = -(r-1);

	verts[2].p_x = 2*r - 1;
	verts[2].p_y = 2*r - 1;
	verts[2].p_z = 0;
	verts[2].c_u = r;
	verts[2].c_v = r;

	
	verts[3].p_x = -1 - 2*(r-1);
	verts[3].p_y = -1 - 2*(r-1);
	verts[3].p_z = 0;
	verts[3].c_u = -(r-1);
	verts[3].c_v = -(r-1);

	verts[4].p_x = 2*r - 1;
	verts[4].p_y = 2*r - 1;
	verts[4].p_z = 0;
	verts[4].c_u = r;
	verts[4].c_v = r;

	verts[5].p_x = -1 - 2*(r-1);
	verts[5].p_y = 2*r - 1;
	verts[5].p_z = 0;
	verts[5].c_u = -(r-1);
	verts[5].c_v = r;

	g.mesh_vb_size = 6;

	glGenBuffers(1, &g.mesh_vb);
	glBindBuffer(GL_ARRAY_BUFFER, g.mesh_vb);
	glBufferData(GL_ARRAY_BUFFER, g.mesh_vb_size * sizeof(VertexData), verts, GL_STATIC_DRAW);

	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void*)0);
	glEnableVertexAttribArray(0);

	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void*)12);
	glEnableVertexAttribArray(1);
}

void draw_mesh(unsigned int shaderprogram, int nfuncs, int *indices, float *coeffs, int *indices_low, float *coeffs_low, bool use_linear)
{
	// bind textures
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, g.textures[g.texture_used]);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	if (use_linear)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
	}

	// bind and run program
	glUseProgram(shaderprogram);
	int loc;

	if (indices)
	{
		loc = glGetUniformLocation(shaderprogram, "indices");
		glUniform3iv(loc, 6*nfuncs, indices);
	}

	if (coeffs)
	{
		loc = glGetUniformLocation(shaderprogram, "coeffs");
		glUniform4fv(loc, 6*nfuncs, coeffs);
	}

	if (indices_low)
	{
		loc = glGetUniformLocation(shaderprogram, "indices_low");
		glUniform3iv(loc, 6*nfuncs, indices_low);
	}

	if (coeffs_low)
	{
		loc = glGetUniformLocation(shaderprogram, "coeffs_low");
		glUniform4fv(loc, 6*nfuncs, coeffs_low);
	}

	loc = glGetUniformLocation(shaderprogram, "un_tex");
	glUniform1i(loc, 0); // use texture unit 0

	loc = glGetUniformLocation(shaderprogram, "mvp");
	glUniformMatrix4fv(loc, 1, 1, g.mvp.m);

	glBindBuffer(GL_ARRAY_BUFFER, g.mesh_vb);

	for (int i = 0; i < iters; i++)
		glDrawArrays(GL_TRIANGLES, 0, g.mesh_vb_size);
}

void draw() 
{
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	if (filter_names[filter_mode] == "hw trilinear")
		draw_mesh(g.shader_hw.shaderprogram, 0, 0, 0, 0, 0, true);
	else if (filter_names[filter_mode] == "sw trilinear")
		draw_mesh(g.shader_trilinear.shaderprogram, 0, 0, 0, 0, 0, false);





	else if (filter_names[filter_mode] == "feline")
		draw_mesh(g.shader_feline.shaderprogram, 0, 0, 0, 0, 0, true);
	
	else if (filter_names[filter_mode] == "feline_sample8")
		draw_mesh(g.shader_feline_sample8.shaderprogram, 
		8, 
		&indices_gaussian_8_lin[0][0][0], 
		&coeffs_gaussian_8_lin[0][0], 
		&indices_gaussian_low_8_lin[0][0][0], 
		&coeffs_gaussian_low_8_lin[0][0], false);
	
	//else if (filter_mode == 4)
	//{
	//	unsigned int shaderprogram = g.shader_aniso_sample8.shaderprogram;
	//	// bind textures
	//	glActiveTexture(GL_TEXTURE0);
	//	glBindTexture(GL_TEXTURE_2D, g.textures[g.texture_used]);
	//	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	//	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	//	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	//	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
	//	// bind and run program
	//	glUseProgram(shaderprogram);
	//	int loc;

	//	loc = glGetUniformLocation(shaderprogram, "indices");
	//	glUniform3iv(loc, 6*8, &indices_gaussian_8_lin[0][0][0]);

	//	loc = glGetUniformLocation(shaderprogram, "coeffs");
	//	glUniform4fv(loc, 6*8, &coeffs_gaussian_8_lin[0][0]);
	//	
	//	loc = glGetUniformLocation(shaderprogram, "indices_low");
	//	glUniform3iv(loc, 6*8, &indices_gaussian_low_8_lin[0][0][0]);

	//	loc = glGetUniformLocation(shaderprogram, "coeffs_low");
	//	glUniform4fv(loc, 6*8, &coeffs_gaussian_low_8_lin[0][0]);

	//	loc = glGetUniformLocation(shaderprogram, "aniso_param");
	//	glUniform2fv(loc, 108, &ansio_guassian_16[0][0][0]);

	//	loc = glGetUniformLocation(shaderprogram, "un_tex");
	//	glUniform1i(loc, 0); // use texture unit 0

	//	loc = glGetUniformLocation(shaderprogram, "mvp");
	//	glUniformMatrix4fv(loc, 1, 1, g.mvp.m);

	//	glBindBuffer(GL_ARRAY_BUFFER, g.mesh_vb);

	//	for (int i = 0; i < iters; i++)
	//		glDrawArrays(GL_TRIANGLES, 0, g.mesh_vb_size);
	//}



	// tent
	else if (filter_names[filter_mode] == "tent_2_sym_uniform")
	{
		draw_mesh(g.shader_sample2_lin_sym_uniform.shaderprogram, 
			      2, 
				  &indices_tent_2_lin[0][0][0], 
				  &coeffs_tent_2_lin[0][0], 
		          &indices_tent_low_2_lin[0][0][0], 
			      &coeffs_tent_low_2_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_3_sym_uniform")
	{
		draw_mesh(g.shader_sample3_lin_sym_uniform.shaderprogram, 
			      3, 
				  &indices_tent_3_lin[0][0][0], 
				  &coeffs_tent_3_lin[0][0], 
		          &indices_tent_low_3_lin[0][0][0], 
			      &coeffs_tent_low_3_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_4_sym_uniform")
	{
		draw_mesh(g.shader_sample4_lin_sym_uniform.shaderprogram, 
			      4, 
				  &indices_tent_4_lin[0][0][0], 
				  &coeffs_tent_4_lin[0][0], 
		          &indices_tent_low_4_lin[0][0][0], 
			      &coeffs_tent_low_4_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_5_sym_uniform")
	{
		draw_mesh(g.shader_sample5_lin_sym_uniform.shaderprogram, 
			      5, 
				  &indices_tent_5_lin[0][0][0], 
				  &coeffs_tent_5_lin[0][0], 
		          &indices_tent_low_5_lin[0][0][0], 
			      &coeffs_tent_low_5_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_6_sym_uniform")
	{
		draw_mesh(g.shader_sample6_lin_sym_uniform.shaderprogram, 
			      6, 
				  &indices_tent_6_lin[0][0][0], 
				  &coeffs_tent_6_lin[0][0], 
		          &indices_tent_low_6_lin[0][0][0], 
			      &coeffs_tent_low_6_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_7_sym_uniform")
	{
		draw_mesh(g.shader_sample7_lin_sym_uniform.shaderprogram, 
			      7, 
				  &indices_tent_7_lin[0][0][0], 
				  &coeffs_tent_7_lin[0][0], 
		          &indices_tent_low_7_lin[0][0][0], 
			      &coeffs_tent_low_7_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_8_sym_uniform")
	{
		draw_mesh(g.shader_sample8_lin_sym_uniform.shaderprogram, 
			      8, 
				  &indices_tent_8_lin[0][0][0], 
				  &coeffs_tent_8_lin[0][0], 
		          &indices_tent_low_8_lin[0][0][0], 
			      &coeffs_tent_low_8_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_9_sym_uniform")
	{
		draw_mesh(g.shader_sample9_lin_sym_uniform.shaderprogram, 
			      9, 
				  &indices_tent_9_lin[0][0][0], 
				  &coeffs_tent_9_lin[0][0], 
		          &indices_tent_low_9_lin[0][0][0], 
			      &coeffs_tent_low_9_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "tent_10_sym_uniform")
	{
		draw_mesh(g.shader_sample10_lin_sym_uniform.shaderprogram, 
			      10, 
				  &indices_tent_10_lin[0][0][0], 
				  &coeffs_tent_10_lin[0][0], 
		          &indices_tent_low_10_lin[0][0][0], 
			      &coeffs_tent_low_10_lin[0][0], false);
	}


	
	// lanczos2
	else if (filter_names[filter_mode] == "lanczos2_2_sym_uniform")
	{
		draw_mesh(g.shader_sample2_lin_sym_uniform.shaderprogram, 
			      2, 
				  &indices_lanczos2_2_lin[0][0][0], 
				  &coeffs_lanczos2_2_lin[0][0], 
		          &indices_lanczos2_low_2_lin[0][0][0], 
			      &coeffs_lanczos2_low_2_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_3_sym_uniform")
	{
		draw_mesh(g.shader_sample3_lin_sym_uniform.shaderprogram, 
			      3, 
				  &indices_lanczos2_3_lin[0][0][0], 
				  &coeffs_lanczos2_3_lin[0][0], 
		          &indices_lanczos2_low_3_lin[0][0][0], 
			      &coeffs_lanczos2_low_3_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_4_sym_uniform")
	{
		draw_mesh(g.shader_sample4_lin_sym_uniform.shaderprogram, 
			      4, 
				  &indices_lanczos2_4_lin[0][0][0], 
				  &coeffs_lanczos2_4_lin[0][0], 
		          &indices_lanczos2_low_4_lin[0][0][0], 
			      &coeffs_lanczos2_low_4_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_5_sym_uniform")
	{
		draw_mesh(g.shader_sample5_lin_sym_uniform.shaderprogram, 
			      5, 
				  &indices_lanczos2_5_lin[0][0][0], 
				  &coeffs_lanczos2_5_lin[0][0], 
		          &indices_lanczos2_low_5_lin[0][0][0], 
			      &coeffs_lanczos2_low_5_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_6_sym_uniform")
	{
		draw_mesh(g.shader_sample6_lin_sym_uniform.shaderprogram, 
			      6, 
				  &indices_lanczos2_6_lin[0][0][0], 
				  &coeffs_lanczos2_6_lin[0][0], 
		          &indices_lanczos2_low_6_lin[0][0][0], 
			      &coeffs_lanczos2_low_6_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_7_sym_uniform")
	{
		draw_mesh(g.shader_sample7_lin_sym_uniform.shaderprogram, 
			      7, 
				  &indices_lanczos2_7_lin[0][0][0], 
				  &coeffs_lanczos2_7_lin[0][0], 
		          &indices_lanczos2_low_7_lin[0][0][0], 
			      &coeffs_lanczos2_low_7_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_8_sym_uniform")
	{
		draw_mesh(g.shader_sample8_lin_sym_uniform.shaderprogram, 
			      8, 
				  &indices_lanczos2_8_lin[0][0][0], 
				  &coeffs_lanczos2_8_lin[0][0], 
		          &indices_lanczos2_low_8_lin[0][0][0], 
			      &coeffs_lanczos2_low_8_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_9_sym_uniform")
	{
		draw_mesh(g.shader_sample9_lin_sym_uniform.shaderprogram, 
			      9, 
				  &indices_lanczos2_9_lin[0][0][0], 
				  &coeffs_lanczos2_9_lin[0][0], 
		          &indices_lanczos2_low_9_lin[0][0][0], 
			      &coeffs_lanczos2_low_9_lin[0][0], false);
	}
	else if (filter_names[filter_mode] == "lanczos2_10_sym_uniform")
	{
		draw_mesh(g.shader_sample10_lin_sym_uniform.shaderprogram, 
			      10, 
				  &indices_lanczos2_10_lin[0][0][0], 
				  &coeffs_lanczos2_10_lin[0][0], 
		          &indices_lanczos2_low_10_lin[0][0][0], 
			      &coeffs_lanczos2_low_10_lin[0][0], false);
	}

	// gaussian
	else if (filter_names[filter_mode] == "gaussian_8_sym_uniform")
	{
		draw_mesh(g.shader_sample8_lin_sym_uniform.shaderprogram, 
			      8, 
				  &indices_gaussian_8_lin[0][0][0], 
				  &coeffs_gaussian_8_lin[0][0], 
		          &indices_gaussian_low_8_lin[0][0][0], 
			      &coeffs_gaussian_low_8_lin[0][0], false);
	}
	
	
	// box
	else if (filter_names[filter_mode] == "box_4_sym_uniform")
	{
		draw_mesh(g.shader_sample4_lin_sym_uniform.shaderprogram, 
			      4, 
				  &indices_box_4_lin[0][0][0], 
				  &coeffs_box_4_lin[0][0], 
		          &indices_box_low_4_lin[0][0][0], 
			      &coeffs_box_low_4_lin[0][0], false);
	}
	
	else if (filter_names[filter_mode] == "box_8_sym_uniform")
	{
		draw_mesh(g.shader_sample8_lin_sym_uniform.shaderprogram, 
			      8, 
				  &indices_box_8_lin[0][0][0], 
				  &coeffs_box_8_lin[0][0], 
		          &indices_box_low_8_lin[0][0][0], 
			      &coeffs_box_low_8_lin[0][0], false);
	}
}

void input()
{
	bool set_tex = false;

	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		if (event.type == SDL_QUIT)
		{
			g.running = false;
		}
		else if (event.type == SDL_KEYDOWN)
		{
			if (event.key.keysym.sym == SDLK_SPACE)
			{
				filter_mode += 1;
				filter_mode %= filter_names.s;
				set_tex = true;
			}
			else if (event.key.keysym.sym == SDLK_BACKSPACE)
			{
				filter_mode += filter_names.s - 1;
				filter_mode %= filter_names.s;
				set_tex = true;
			}
			
			else if (event.key.keysym.sym == SDLK_LEFTBRACKET)
			{
				g.texture_used += texture_names.s - 1;
				g.texture_used %= texture_names.s;
			}
			else if (event.key.keysym.sym == SDLK_RIGHTBRACKET)
			{
				g.texture_used += 1;
				g.texture_used %= texture_names.s;
			}
			
			else if (event.key.keysym.sym == SDLK_o)
			{
				ifstream f("view.txt");
				if (!f.fail())
				{
					f >> g.trans[0] >> g.trans[1] >> g.trans[2];
					f >> g.rot.m[0] >> g.rot.m[1] >> g.rot.m[2]
					  >> g.rot.m[3] >> g.rot.m[4] >> g.rot.m[5]
					  >> g.rot.m[6] >> g.rot.m[7] >> g.rot.m[8];
					f.close();
				}
			}
			else if (event.key.keysym.sym == SDLK_p)
			{
				FILE *f = fopen("view.txt", "wb");
				if (f)
				{
					fprintf(f, "%f %f %f\n", g.trans[0], g.trans[1], g.trans[2]);
					fprintf(f, "\n");
					fprintf(f, "%f %f %f\n", g.rot.m[0], g.rot.m[1], g.rot.m[2]);
					fprintf(f, "%f %f %f\n", g.rot.m[3], g.rot.m[4], g.rot.m[5]);
					fprintf(f, "%f %f %f\n", g.rot.m[6], g.rot.m[7], g.rot.m[8]);
					fclose(f);
				}
			}
			else if (event.key.keysym.sym == SDLK_m)
			{
				movie.grabbing = true;
			}
		}
		else if (event.type == SDL_MOUSEMOTION)
		{
			if (event.motion.state == 1)
			{
				// trackball rotate
				vect3f yv, xv;
				xv.set(1,0,0);
				yv.set(0,1,0);

				float xa = -event.motion.xrel * .01;
				float ya = -event.motion.yrel * .01;

				g.rot = matrix3f::rotaxis(xv, ya) * matrix3f::rotaxis(yv, xa) * g.rot;
			}
			else if (event.motion.state == 2)
			{
				// zoom
				double ya = -event.motion.yrel * 2.0 / screen_height;

				g.rot *= pow(2.0, ya);
			}
			else if (event.motion.state == 4)
			{
				// translation
				vect3f yv, xv;
				xv.set(1,0,0);
				yv.set(0,1,0);

				float xa = event.motion.xrel * 2.0 / screen_height;
				float ya = -event.motion.yrel * 2.0 / screen_height;

				g.trans += (xv * xa + yv * ya);
			}
		}
	}

	g.trans[2] = -2;
	if (set_tex)
	{
	texture_names.push_back("box"); // 0
	texture_names.push_back("tent"); // 1
	texture_names.push_back("lanczos2"); // 2
	texture_names.push_back("jinc2"); // 3
	texture_names.push_back("gauss"); // 4
		if (filter_mode >= 0 && filter_mode <= 1)
			g.texture_used = 1;
		else if (filter_mode >= 2 && filter_mode <= 3)
			g.texture_used = 4;
		else if (filter_mode >= 4 && filter_mode <= 12)
			g.texture_used = 1;
		else if (filter_mode >= 13 && filter_mode <= 21)
			g.texture_used = 2;
		else if (filter_mode == 22 || filter_mode == 23)
			g.texture_used = 0;
		else if (filter_mode == 24)
			g.texture_used = 4;
	}

	//========= lizard
	/*g.rot.m[0] = 0.951057; // lizard
	g.rot.m[1] = 0.309017;
	g.rot.m[2] = 0;

	g.rot.m[3] = -0.0954915;
	g.rot.m[4] = 0.293893;
	g.rot.m[5] = 0.951057;
	
	g.rot.m[6] = 0.293893;
	g.rot.m[7] = -0.904508;
	g.rot.m[8] = 0.309017;

	g.rot *= (1.0 / .9);

	g.trans.set(-.36, 0, -.4);
	g.trans = g.rot.mult(g.trans);*/


	//========== infinite plane
	//g.trans.set(0,0,-2);
	//vect3f axis_x; axis_x.set(1,0,0);
	//g.rot = matrix3f::rotaxis(axis_x, 3.1415926535 * -.24);
	//g.rot *= .5;

	//=========== animated rotation
	if (movie.grabbing)
	{
		g.trans.set(0,0,2);
		vect3f axis_x; axis_x.set(1,0,0);
		vect3f axis_z; axis_z.set(0,0,1);
		g.rot = matrix3f::rotaxis(axis_x, 3.1415926535 * -.24);
		g.rot = g.rot.mult(matrix3f::rotaxis(axis_z, 3.1415926535/180.0 * .5 * movie.movie.s));
	}


	// build the MVP matrix
	matrix4f modelview, projection;
	modelview.fromRotTrans(g.rot, g.trans);
	projection.perspectiveFOV(.1, 100, 90, (float)screen_width / (float)screen_height);
	g.mvp = projection * modelview;
}

template<class F>
void load_mipmap(MipStack &mip, F func, string type, string srcname, Array2D<vect4f> &src)
{
	int levels = log((double)src.s[0]) / log(2.0) + 1.5;
	mip.img.resize(levels);
	mip.s[0] = src.s[0];
	mip.s[1] = src.s[1];

	// try to load mipmaps
	char buf[1024];
	int loaded = 0;
	for (int i = 0; i < levels; i++)
	{
		sprintf(buf, "raw/%s.%s.mip%d.raw", srcname.c_str(), type.c_str(), i);
		FILE *f = fopen(buf, "rb");
		if (f == 0)
			break;

		Array2D<vect4f> &img = mip.img[i];

		vect2i s;
		fread(s.v, 4, 2, f);
		img.resize(s);

		fread(img.data, 16, img.data_size, f);

		fclose(f);
		loaded++;
		printf("loaded %s %d %d\n", buf, img.s[0], img.s[1]);
	}
	
	if (loaded != levels)
	{
		// calc the mipmap
		mip.img.clear();

		printf("calculating mip_%s...\n", type.c_str());
		func(mip, src);

		// save for future
		assert(mip.img.s == levels);
		for (int i = 0; i < levels; i++)
		{
			sprintf(buf, "raw/%s.%s.mip%d.raw", srcname.c_str(), type.c_str(), i);
			FILE *f = fopen(buf, "wb");
			assert(f);

			Array2D<vect4f> &img = mip.img[i];

			fwrite(img.s, 4, 2, f);

			fwrite(img.data, 16, img.data_size, f);

			fclose(f);
			printf("saved %s %d %d\n", buf, img.s[0], img.s[1]);

			// save png
			Array2D<vect4f> gamma = img;
			for (int k = 0; k < gamma.data_size; k++)
			{
				for (int l = 0; l < 3; l++)
					gamma.data[k][l] = RGB32F_to_sRGB32F(gamma.data[k][l]);
			}

			sprintf(buf, "raw/%s.%s.mip%d.png", srcname.c_str(), type.c_str(), i);
			save_image(buf, gamma, 24);
		}
	}
}

void convert_srgb_to_linear(Array2D<vect4f> &src)
{
#ifdef USE_SRGB
	for (int i = 0; i < src.data_size; i++)
		for (int j = 0; j < 3; j++)
			src.data[i][j] = sRGB32F_to_RGB32F(src.data[i][j]);
#endif
}

int main(int argc, char *argv[])
{
	system("mkdir raw");

	g.texture_used = 1;
	g.trans = 0;
	g.rot.setIdentity();

	texture_names.push_back("box");
	texture_names.push_back("tent");
	texture_names.push_back("lanczos2");
	texture_names.push_back("jinc2");
	texture_names.push_back("gauss");
	texture_names.push_back("rainbow");

	filter_names.push_back("hw trilinear"); // 0
	//filter_names.push_back("hw + 1");
	filter_names.push_back("sw trilinear");
	//filter_names.push_back("hw major");

	filter_names.push_back("feline"); // 2
	filter_names.push_back("feline_sample8");
	//filter_names.push_back("aniso_sample8");

	filter_names.push_back("tent_2_sym_uniform"); //4
	filter_names.push_back("tent_3_sym_uniform");
	filter_names.push_back("tent_4_sym_uniform");
	filter_names.push_back("tent_5_sym_uniform");
	filter_names.push_back("tent_6_sym_uniform");
	filter_names.push_back("tent_7_sym_uniform");
	filter_names.push_back("tent_8_sym_uniform");
	filter_names.push_back("tent_9_sym_uniform");
	filter_names.push_back("tent_10_sym_uniform");

	filter_names.push_back("lanczos2_2_sym_uniform"); // 13
	filter_names.push_back("lanczos2_3_sym_uniform");
	filter_names.push_back("lanczos2_4_sym_uniform");
	filter_names.push_back("lanczos2_5_sym_uniform");
	filter_names.push_back("lanczos2_6_sym_uniform");
	filter_names.push_back("lanczos2_7_sym_uniform");
	filter_names.push_back("lanczos2_8_sym_uniform");
	filter_names.push_back("lanczos2_9_sym_uniform");
	filter_names.push_back("lanczos2_10_sym_uniform");

	filter_names.push_back("box_4_sym_uniform"); // 22
	filter_names.push_back("box_8_sym_uniform");
	
	filter_names.push_back("gaussian_8_sym_uniform"); // 24

	// init code
	SDL_Init(SDL_INIT_VIDEO);
	SDL_Surface *screen = SDL_SetVideoMode(screen_width, screen_height, 32, SDL_HWSURFACE | SDL_OPENGL | SDL_HWACCEL);
	
	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

	glewInit();

	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);
#ifdef USE_SRGB
	glEnable(GL_FRAMEBUFFER_SRGB);
#endif
	
	// get version
	const GLubyte *version_gl = glGetString(GL_VERSION);
	printf("OpenGL %s\n", version_gl);
	const GLubyte *version_glsl = glGetString(GL_SHADING_LANGUAGE_VERSION);
	printf("GLSL %s\n", version_glsl);

	


	// load mesh
	create_quad();
	//load_mesh("lizard_closed_sub3.obj");
	//load_mesh("test.obj");
	//load_mesh("axe.obj");


	// load some data
	//string imgname = "lizard.png";
	//string imgname = "axe.png";
	//string imgname = "green.png";
	//string imgname = "checker10.png";
	string imgname = "lena.png";
	//string imgname = "wallflower.png";
	//string imgname = "fleur_bo.png";
	//string imgname = "radial_2048.png";
	//string imgname = "zone_2048.png";

	printf("loading image\n");
	Array2D<vect4f> src;
	load_into(src, imgname);
	convert_srgb_to_linear(src);
	
	// calc mipmap
	printf("building mipmaps\n");
	MipStack mip_box, mip_tent, mip_lanczos2, mip_jinc2, mip_gauss, mip_rainbow;

	load_mipmap(mip_box, calc_mip_box, "box", imgname, src);
	load_mipmap(mip_tent, calc_mip_tent, "tent", imgname, src);
	load_mipmap(mip_lanczos2, calc_mip_lanczos2, "lanczos2", imgname, src);
	load_mipmap(mip_jinc2, calc_mip_jinc2, "jinc2", imgname, src);
	load_mipmap(mip_gauss, calc_mip_gauss, "gauss", imgname, src);
	load_mipmap(mip_rainbow, calc_mip_rainbow, "rainbow", imgname, src);






	// load shaders
	Array<string> attribs;
	attribs.push_back("in_position");
	attribs.push_back("in_coord");
	g.shader_hw = create_shader("shaders/simple.vert.c", "shaders/hw.frag.c", attribs);
	g.shader_trilinear = create_shader("shaders/simple.vert.c", "shaders/trilinear.frag.c", attribs);
	g.shader_feline = create_shader("shaders/simple.vert.c", "shaders/feline.frag.c", attribs);
	g.shader_feline_sample8 = create_shader("shaders/simple.vert.c", "shaders/feline_sample8.frag.c", attribs);
	g.shader_sample2_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample2_lin_sym_uniform.frag.c", attribs);
	g.shader_sample3_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample3_lin_sym_uniform.frag.c", attribs);
	g.shader_sample4_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample4_lin_sym_uniform.frag.c", attribs);
	g.shader_sample5_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample5_lin_sym_uniform.frag.c", attribs);
	g.shader_sample6_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample6_lin_sym_uniform.frag.c", attribs);
	g.shader_sample7_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample7_lin_sym_uniform.frag.c", attribs);
	g.shader_sample8_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample8_lin_sym_uniform.frag.c", attribs);
	g.shader_sample9_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample9_lin_sym_uniform.frag.c", attribs);
	g.shader_sample10_lin_sym_uniform = create_shader("shaders/simple.vert.c", "shaders/sample10_lin_sym_uniform.frag.c", attribs);


	printf("upload mipmaps\n");
	// upload mipmap
	glGenTextures(1, &g.tex_box);
	glBindTexture(GL_TEXTURE_2D, g.tex_box);

	for (int i = 0; i < mip_box.img.s; i++)
		glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA32F, mip_box.img[i].s[0], mip_box.img[i].s[1], 0, GL_RGBA, GL_FLOAT, mip_box.img[i].data);

	// upload mipmap
	glGenTextures(1, &g.tex_tent);
	glBindTexture(GL_TEXTURE_2D, g.tex_tent);

	for (int i = 0; i < mip_tent.img.s; i++)
		glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA32F, mip_tent.img[i].s[0], mip_tent.img[i].s[1], 0, GL_RGBA, GL_FLOAT, mip_tent.img[i].data);

	// upload mipmap
	glGenTextures(1, &g.tex_lanczos2);
	glBindTexture(GL_TEXTURE_2D, g.tex_lanczos2);

	for (int i = 0; i < mip_lanczos2.img.s; i++)
		glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA32F, mip_lanczos2.img[i].s[0], mip_lanczos2.img[i].s[1], 0, GL_RGBA, GL_FLOAT, mip_lanczos2.img[i].data);

	// upload mipmap
	glGenTextures(1, &g.tex_jinc2);
	glBindTexture(GL_TEXTURE_2D, g.tex_jinc2);

	for (int i = 0; i < mip_jinc2.img.s; i++)
		glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA32F, mip_jinc2.img[i].s[0], mip_jinc2.img[i].s[1], 0, GL_RGBA, GL_FLOAT, mip_jinc2.img[i].data);

	// upload mipmap
	glGenTextures(1, &g.tex_gauss);
	glBindTexture(GL_TEXTURE_2D, g.tex_gauss);

	for (int i = 0; i < mip_gauss.img.s; i++)
		glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA32F, mip_gauss.img[i].s[0], mip_gauss.img[i].s[1], 0, GL_RGBA, GL_FLOAT, mip_gauss.img[i].data);

	// upload mipmap
	glGenTextures(1, &g.tex_rainbow);
	glBindTexture(GL_TEXTURE_2D, g.tex_rainbow);

	for (int i = 0; i < mip_rainbow.img.s; i++)
		glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA32F, mip_rainbow.img[i].s[0], mip_rainbow.img[i].s[1], 0, GL_RGBA, GL_FLOAT, mip_rainbow.img[i].data);

	// main loop
	printf("begin render loop\n");
	double time_prev, time_current;
	time_prev = time_current = get_time();

	int frame = 0;
	while (g.running)
	{
		time_current = get_time();
		if (time_current - time_prev >= .2)
		{
			printf("fps = %f filter = %s texture = %s    \r", frame / (time_current - time_prev), filter_names[filter_mode].c_str(), texture_names[g.texture_used].c_str());
			time_prev = time_current;
			frame = 0;
		}

		input();
		draw();
		frame++;

		if (movie.grabbing)
		{
			movie.grabMovieFrame();
			if (movie.movie.s == 180)
			{
				movie.saveMovieToDir("movie");
				movie.movie.clear();
				movie.grabbing = false;
			}
		}

		SDL_GL_SwapBuffers();
	}


	// cleanup code
	SDL_Quit();
	return 0;
}
