#include "mipsample.h"

static double pi = 3.14159265358979323846264338327950288;

static double l2(double x)
{
	if (x <= -2 || x >= 2)
		return 0;
	else if (fabs(x) < .0001)
		return 1;
	else
		return 2*sin(pi*x/2) * sin(pi*x) / (pi*pi*x*x);
}

static double filter(double x, double y)
{
	return l2(x)*l2(y);
}

vect4f sample_lanczos2_ideal(Array2D<vect4f> &src, float u, float v, float w)
{
	float scale = pow(2, w);
	float x = src.s[0] * u - .5;
	float y = src.s[1] * v - .5;

	int mini = ceil(x - 2*scale);
	int maxi = floor(x + 2*scale);

	int minj = ceil(y - 2*scale);
	int maxj = floor(y + 2*scale);

	double sum_weight = 0;
	vect4f sum;
	sum = 0;
	for (int j = minj; j <= maxj; j++)
	{
		for (int i = mini; i <= maxi; i++)
		{
			double weight = filter((i-x)/scale, (j-y)/scale);
#ifdef MIP_WRAP
			sum += src.get_wrap(i, j) * weight;
#else
			sum += src.get_clamp(i, j) * weight;
#endif
			sum_weight += weight;
		}
	}
	sum /= sum_weight;

	return sum;
}
