#include "mipsample.h"

static double tent(double x)
{
	if (x <= -1 || x >= 1)
		return 0;
	else if (x < 0)
		return 1 + x;
	else
		return 1 - x;
}

static double filter(double x, double y)
{
	return tent(x)*tent(y);
}

vect4f sample_tent_ideal(Array2D<vect4f> &src, float u, float v, float w)
{
	float scale = pow(2, w);
	float x = src.s[0] * u - .5;
	float y = src.s[1] * v - .5;

	double r = 1;

	int mini = ceil(x - r*scale);
	int maxi = floor(x + r*scale);

	int minj = ceil(y - r*scale);
	int maxj = floor(y + r*scale);

	double sum_weight = 0;
	vect4f sum;
	sum = 0;
	for (int j = minj; j <= maxj; j++)
	{
		for (int i = mini; i <= maxi; i++)
		{
			double weight = filter((i-x)/scale, (j-y)/scale);
#ifdef MIP_WRAP
			sum += src.get_wrap(i, j) * weight;
#else
			sum += src.get_clamp(i, j) * weight;
#endif
			sum_weight += weight;
		}
	}
	sum /= sum_weight;

	return sum;
}
