#include "shader.h"
#include "array.h"
#include "mmap_file.h"
#include <GL/glew.h>
#include <math.h>
#include <assert.h>
#include <string>

using namespace std;

ShaderObject create_shader(string fn_vert, string fn_frag, Array<string> &attribs)
{
	return create_shader(fn_vert, fn_frag, "", attribs);
}

ShaderObject create_shader(string fn_vert, string fn_frag, string fn_geom, Array<string> &attribs)
{
	ShaderObject so;

	// create shader program
	so.shaderprogram = glCreateProgram();

	// create vertex shader
	string vertexsourcestr;
	load_file_to_string(fn_vert, vertexsourcestr);
	const GLchar *vertexsource = vertexsourcestr.c_str();
	so.vertexshader = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(so.vertexshader, 1, (const GLchar**)&vertexsource, 0);
	glCompileShader(so.vertexshader);
	int IsCompiled_VS;
	glGetShaderiv(so.vertexshader, GL_COMPILE_STATUS, &IsCompiled_VS);
	if (!IsCompiled_VS)
	{
		int maxLength;
		glGetShaderiv(so.vertexshader, GL_INFO_LOG_LENGTH, &maxLength);
		char *vertexInfoLog = new char [maxLength];
		glGetShaderInfoLog(so.vertexshader, maxLength, &maxLength, vertexInfoLog);
		printf("Error: Compiling vertex shader %s\n\n%s\n", fn_vert.c_str(), vertexInfoLog);
		delete[] vertexInfoLog;
		exit(1);
	}
	glAttachShader(so.shaderprogram, so.vertexshader);

	// create fragment shader
	string fragmentsourcestr;
	load_file_to_string(fn_frag, fragmentsourcestr);
	const GLchar *fragmentsource = fragmentsourcestr.c_str();
	so.fragmentshader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(so.fragmentshader, 1, (const GLchar**)&fragmentsource, 0);
	glCompileShader(so.fragmentshader);
	int IsCompiled_FS;
	glGetShaderiv(so.fragmentshader, GL_COMPILE_STATUS, &IsCompiled_FS);
	if (!IsCompiled_FS)
	{
		int maxLength;
		glGetShaderiv(so.fragmentshader, GL_INFO_LOG_LENGTH, &maxLength);
		char *fragmentInfoLog = new char [maxLength];
		glGetShaderInfoLog(so.fragmentshader, maxLength, &maxLength, fragmentInfoLog);
		printf("Error: Compiling fragment shader %s\n\n%s\n", fn_frag.c_str(), fragmentInfoLog);
		delete[] fragmentInfoLog;
		exit(1);
	}
	glAttachShader(so.shaderprogram, so.fragmentshader);

	// create geometry shader
	if (fn_geom.size())
	{
		string geomsourcestr;
		load_file_to_string(fn_geom, geomsourcestr);
		const GLchar *geomsource = geomsourcestr.c_str();
		so.geometryshader = glCreateShader(GL_GEOMETRY_SHADER);
		glShaderSource(so.geometryshader, 1, (const GLchar**)&geomsource, 0);
		glCompileShader(so.geometryshader);
		int IsCompiled_GS;
		glGetShaderiv(so.geometryshader, GL_COMPILE_STATUS, &IsCompiled_GS);
		if (!IsCompiled_GS)
		{
			int maxLength;
			glGetShaderiv(so.geometryshader, GL_INFO_LOG_LENGTH, &maxLength);
			char *geomInfoLog = new char [maxLength];
			glGetShaderInfoLog(so.geometryshader, maxLength, &maxLength, geomInfoLog);
			printf("Error: Compiling geometry shader %s\n\n%s\n", fn_geom.c_str(), geomInfoLog);
			delete[] geomInfoLog;
			exit(1);
		}
		glAttachShader(so.shaderprogram, so.geometryshader);
	}

	// link shader program
	for (int i = 0; i < attribs.s; i++)
		glBindAttribLocation(so.shaderprogram, i, attribs[i].c_str());

	glLinkProgram(so.shaderprogram);
	int IsLinked;
	glGetProgramiv(so.shaderprogram, GL_LINK_STATUS, (int *)&IsLinked);
	if (!IsLinked)
	{
		int maxLength;
		glGetProgramiv(so.shaderprogram, GL_INFO_LOG_LENGTH, &maxLength);
		char *shaderProgramInfoLog = new char [maxLength];
		glGetProgramInfoLog(so.shaderprogram, maxLength, &maxLength, shaderProgramInfoLog);
		printf("Error: Linking shader program %s %s %s\n\n%s\n", fn_vert.c_str(), fn_geom.c_str(), fn_frag.c_str(), shaderProgramInfoLog);
		delete[] shaderProgramInfoLog;
		exit(1);
	}

	return so;
}
