#version 410

const int max_probes = 16;

uniform sampler2D un_tex;

in vec2 ex_coord;

void main()
{
	ivec2 ts = textureSize(un_tex, 0);
	float ts_inv = 1.0 / ts.x;
	int mipmax = int(log2(ts.x));

	//==================================================
	// calculate jacobian of screen to texture map
	//==================================================

	vec2 dx = dFdx(ex_coord);
	vec2 dy = dFdy(ex_coord);
	
	float dudx = dx.x;
	float dudy = dy.x;
	float dvdx = dx.y;
	float dvdy = dy.y;

	//==================================================
	// calculate ellipse properties
	//==================================================

	// Add 1 (ts_inv^2) to a and c from page 7 feline -> page 60 Heckbert thesis.
	// I'm not convinced this is a good idea.
	//float a = dvdx*dvdx + dvdy*dvdy;
	float a = dvdx*dvdx + dvdy*dvdy + ts_inv*ts_inv; 
	float b = -2*(dudx*dvdx + dudy*dvdy);
	//float c = dudx*dudx + dudy*dudy;
	float c = dudx*dudx + dudy*dudy + ts_inv*ts_inv; 
	float f = a*c - b*b*.25;
	a /= f;
	b /= f;
	c /= f;
	float root = sqrt((a-c)*(a-c)+b*b);
	
	float minor = max(ts_inv, inversesqrt(.5*(a+c+root)));
	float major = max(ts_inv, inversesqrt(.5*(a+c-root)));
	
	float theta = atan(b/(a-c))*.5;
	if (a > c)
		theta += 3.1415926535897 * .5;
	

	//==================================================
	// calculate the feline sampling parameters
	//==================================================
	float fprobes = 2 * major / minor - 1;
	int iprobes = int(fprobes + .5);
	
	/* 
	// only increase filter radius when not enough probes
	if (iprobes > max_probes)
	{
		iprobes = max_probes;
		minor = 2*major / (iprobes+1);
	}
	*/

	///*
	// exact feline implementation
	iprobes = min(iprobes, max_probes);
	if (iprobes < fprobes)
		minor = 2*major / (iprobes+1);
	//*/
	
	float ex_z = log2(minor * ts.x);
	if (ex_z >= mipmax)
		iprobes = 1;
	

	float lineLength = 2 * ( major - minor );
	float lineLength_iprobes;
	if (iprobes <= 1)
		lineLength_iprobes = 1;
	else
		lineLength_iprobes = lineLength / ( iprobes - 1.0 );


	vec2 dir;
	dir.x = cos(theta) * lineLength_iprobes;
	dir.y = sin(theta) * lineLength_iprobes;

	//==================================================
	// sample the texture
	//==================================================

	float weight = 0;
	vec3 color = vec3(0,0,0);
	for ( int n = 1 - iprobes; n <= iprobes - 1; n += 2 )
	{
		vec2 pos = ex_coord + dir * n * 0.5;
		float d = ( n * 0.5 * lineLength_iprobes ) / major;
		float w = exp ( -2 * d * d );

		color += w*textureLod(un_tex, pos, ex_z).rgb;
		weight += w;
	}
	gl_FragColor.rgb = color * (1.0 / weight);

	//gl_FragColor.rgb = vec3 ( iprobes & 0x04, iprobes & 0x02, iprobes & 0x01 ); // number of probes
}
