#version 410

const int nfuncs = 8;
const int max_probes = 16;

uniform sampler2D un_tex;
uniform ivec3 indices[6*nfuncs];
uniform vec4 coeffs[6*nfuncs];
uniform ivec3 indices_low[6*nfuncs];
uniform vec4 coeffs_low[6*nfuncs];

in vec2 ex_coord;

int res;
float ex_z;
ivec3 posi;

ivec2 ts;
int mipmax;

vec3 sample_full(vec2 tex_coord, float tex_z)
{
	vec3 sum = vec3(0,0,0);
	ivec3 p;
	vec3 posf, f;
	
	posf.z = exp2(posi.z - tex_z) - 1;
	
	posf.xy = tex_coord * res;
	posi.x = int(floor(posf.x));
	posi.y = int(floor(posf.y));
	
	p.x = int(floor(4*(posf.x - posi.x))); // which part of the domain
	p.y = int(floor(4*(posf.y - posi.y)));
	p.z = int(floor(2*posf.z));
	
	f.x = 4*(posf.x - posi.x) - p.x; // fractional value in [0,1] of point in part
	f.y = 4*(posf.y - posi.y) - p.y;
	f.z = 2*posf.z - p.z;
	
	// convert to symmetric part
	bool flipx = false, flipy = false, flipxy = false;

	if (p.x >= 2)
	{
		flipx = true;
		p.x = 3 - p.x;
		f.x = 1 - f.x;
	}
	if (p.y >= 2)
	{
		flipy = true;
		p.y = 3 - p.y;
		f.y = 1 - f.y;
	}
	if (p.y > p.x)
	{
		flipxy = true;
		p.xy = p.yx;
		f.xy = f.yx;
	}

	int sind = (p.x + p.y + 3*p.z) * nfuncs;
	
	for (int i = 0; i < nfuncs; i++)
	{
		ivec3 o = indices[sind+i];
		
		int exp2oz = (1<<o.z);
		if (flipxy)
			o.xy = o.yx;
		
		if (flipx)
			o.x = (1<<o.z) - 1 - o.x;
		if (flipy)
			o.y = (1<<o.z) - 1 - o.y;

		o.xy += posi.xy*exp2oz;
		o.z = posi.z - o.z;

		vec4 cvec = coeffs[sind+i];
		float c = dot(cvec, vec4(1, f.x, f.y, f.z));
		
		sum += c * textureLod(un_tex, (vec2(o.xy)+.5)/(res*exp2oz), o.z).rgb;
	}

	return sum;
}


vec3 sample_low(vec2 tex_coord, float tex_z)
{
	vec3 sum = vec3(0,0,0);
	ivec3 p;
	vec3 posf, f;
	
	posf.z = exp2(posi.z - tex_z) - 1;
	
	posf.xy = tex_coord * res;
	posi.x = int(floor(posf.x));
	posi.y = int(floor(posf.y));
	
	p.x = int(floor(4*(posf.x - posi.x))); // which part of the domain
	p.y = int(floor(4*(posf.y - posi.y)));
	p.z = int(floor(2*posf.z));
	
	f.x = 4*(posf.x - posi.x) - p.x; // fractional value in [0,1] of point in part
	f.y = 4*(posf.y - posi.y) - p.y;
	f.z = 2*posf.z - p.z;
	
	// convert to symmetric part
	bool flipx = false, flipy = false, flipxy = false;

	if (p.x >= 2)
	{
		flipx = true;
		p.x = 3 - p.x;
		f.x = 1 - f.x;
	}
	if (p.y >= 2)
	{
		flipy = true;
		p.y = 3 - p.y;
		f.y = 1 - f.y;
	}
	if (p.y > p.x)
	{
		flipxy = true;
		p.xy = p.yx;
		f.xy = f.yx;
	}

	int sind = (p.x + p.y + 3*p.z) * nfuncs;
	
	for (int i = 0; i < nfuncs; i++)
	{
		ivec3 o = indices_low[sind+i];
		
		int exp2oz = (1<<o.z);
		if (flipxy)
			o.xy = o.yx;
		
		if (flipx)
			o.x = (1<<o.z) - 1 - o.x;
		if (flipy)
			o.y = (1<<o.z) - 1 - o.y;

		o.xy += posi.xy*exp2oz;
		o.z = posi.z - o.z;

		vec4 cvec = coeffs_low[sind+i];
		float c = dot(cvec, vec4(1, f.x, f.y, f.z));
		
		sum += c * textureLod(un_tex, (vec2(o.xy)+.5)/(res*exp2oz), o.z).rgb;
	}

	return sum;
}


vec3 bilinear(vec2 pin, ivec2 size, int lev)
{
	vec2 p, t, f;
	
	for (int i = 0; i < 2; i++)
	{
		p[i] = pin[i] * size[i] - .5;
		f[i] = floor(p[i]);
		t[i] = p[i] - f[i];
	}
	
	float one = 1.0 / size[0];
	
	f += .5;
	f /= size[0];
	
	return  mix(
			mix(textureLod(un_tex, vec2(f.x,f.y), lev).rgb, textureLod(un_tex, vec2(f.x,f.y+one), lev).rgb, t.y),
			mix(textureLod(un_tex, vec2(f.x+one,f.y), lev).rgb, textureLod(un_tex, vec2(f.x+one,f.y+one), lev).rgb, t.y),
			t.x);
}

vec3 sample8(vec2 tex_coord, float tex_z)
{
	posi.z = int(ceil(tex_z)); // index of lowest resolution
	res = textureSize(un_tex, int(posi.z)).x; // lowest resolution
	
	if (tex_z >= mipmax)
		return textureLod(un_tex, tex_coord, tex_z).rgb;
	else if (tex_z <= 0)
		return bilinear(tex_coord, ts, 0);
	else if (tex_z <= 1)
		return sample_low(tex_coord, tex_z);
	else
		return sample_full(tex_coord, tex_z);
}

void main()
{
	ts = textureSize(un_tex, 0);
	float ts_inv = 1.0 / ts.x;
	mipmax = int(log2(ts.x));

	//==================================================
	// calculate jacobian of screen to texture map
	//==================================================
	vec2 dx = dFdx(ex_coord);
	vec2 dy = dFdy(ex_coord);
	
	float dudx = dx.x;
	float dudy = dy.x;
	float dvdx = dx.y;
	float dvdy = dy.y;

	//==================================================
	// calculate ellipse properties
	//==================================================

	// Add 1 (ts_inv^2) to a and c from page 7 feline -> page 60 Heckbert thesis.
	// I'm not convinced this is a good idea.
	//float a = dvdx*dvdx + dvdy*dvdy;
	float a = dvdx*dvdx + dvdy*dvdy + ts_inv*ts_inv; 
	float b = -2*(dudx*dvdx + dudy*dvdy);
	//float c = dudx*dudx + dudy*dudy;
	float c = dudx*dudx + dudy*dudy + ts_inv*ts_inv; 
	float f = a*c - b*b*.25;
	a /= f;
	b /= f;
	c /= f;
	float root = sqrt((a-c)*(a-c)+b*b);
	
	float minor = max(ts_inv, inversesqrt(.5*(a+c+root)));
	float major = max(ts_inv, inversesqrt(.5*(a+c-root)));
	
	float theta = atan(b/(a-c))*.5;
	if (a > c)
		theta += 3.1415926535897 * .5;
	

	//==================================================
	// calculate the feline sampling parameters
	//==================================================
	float fprobes = 2 * major / minor - 1;
	int iprobes = int(fprobes + .5);
	
	/* 
	// only increase filter radius when not enough probes
	if (iprobes > max_probes)
	{
		iprobes = max_probes;
		minor = 2*major / (iprobes+1);
	}
	*/

	///*
	// exact feline implementation
	iprobes = min(iprobes, max_probes);
	if (iprobes < fprobes)
		minor = 2*major / (iprobes+1);
	//*/
	
	float ex_z = log2(minor * ts.x);
	if (ex_z >= mipmax)
		iprobes = 1;
	

	float lineLength = 2 * ( major - minor );
	float lineLength_iprobes;
	if (iprobes <= 1)
		lineLength_iprobes = 1;
	else
		lineLength_iprobes = lineLength / ( iprobes - 1.0 );


	vec2 dir;
	dir.x = cos(theta) * lineLength_iprobes;
	dir.y = sin(theta) * lineLength_iprobes;

	//==================================================
	// sample the texture
	//==================================================

	float weight = 0;
	vec3 color = vec3(0,0,0);
	for ( int n = 1 - iprobes; n <= iprobes - 1; n += 2 )
	{
		vec2 pos = ex_coord + dir * n * 0.5;
		float d = ( n * 0.5 * lineLength_iprobes ) / major;
		float w = exp ( -2 * d * d );

		color += w*sample8(pos, ex_z).rgb;
		weight += w;
	}
	gl_FragColor.rgb = color * (1.0 / weight);

	/*ex_z = textureQueryLod(un_tex, ex_coord).y;
	gl_FragColor.rgb = sample8(ex_coord, ex_z);*/
}
