#version 410

const int nfuncs = 6;

uniform sampler2D un_tex;
uniform ivec3 indices[6*nfuncs];
uniform vec4 coeffs[6*nfuncs];
uniform ivec3 indices_low[6*nfuncs];
uniform vec4 coeffs_low[6*nfuncs];

in vec2 ex_coord;

int res;
float ex_z;
ivec3 posi;

vec3 sample_full()
{
	vec3 sum = vec3(0,0,0);
	ivec3 p;
	vec3 posf, f;
	
	posf.z = exp2(posi.z - ex_z) - 1;
	
	posf.xy = ex_coord * res;
	posi.x = int(floor(posf.x));
	posi.y = int(floor(posf.y));
	
	p.x = int(floor(4*(posf.x - posi.x))); // which part of the domain
	p.y = int(floor(4*(posf.y - posi.y)));
	p.z = int(floor(2*posf.z));
	
	f.x = 4*(posf.x - posi.x) - p.x; // fractional value in [0,1] of point in part
	f.y = 4*(posf.y - posi.y) - p.y;
	f.z = 2*posf.z - p.z;
	
	// convert to symmetric part
	bool flipx = false, flipy = false, flipxy = false;

	if (p.x >= 2)
	{
		flipx = true;
		p.x = 3 - p.x;
		f.x = 1 - f.x;
	}
	if (p.y >= 2)
	{
		flipy = true;
		p.y = 3 - p.y;
		f.y = 1 - f.y;
	}
	if (p.y > p.x)
	{
		flipxy = true;
		p.xy = p.yx;
		f.xy = f.yx;
	}

	int sind = (p.x + p.y + 3*p.z) * nfuncs;
	
	for (int i = 0; i < nfuncs; i++)
	{
		ivec3 o = indices[sind+i];
		
		int exp2oz = (1<<o.z);
		if (flipxy)
			o.xy = o.yx;
		
		if (flipx)
			o.x = (1<<o.z) - 1 - o.x;
		if (flipy)
			o.y = (1<<o.z) - 1 - o.y;

		o.xy += posi.xy*exp2oz;
		o.z = posi.z - o.z;

		vec4 cvec = coeffs[sind+i];
		float c = dot(cvec, vec4(1, f.x, f.y, f.z));
		
		sum += c * textureLod(un_tex, (vec2(o.xy)+.5)/(res*exp2oz), o.z).rgb;
	}

	return sum;
}


vec3 sample_low()
{
	vec3 sum = vec3(0,0,0);
	ivec3 p;
	vec3 posf, f;
	
	posf.z = exp2(posi.z - ex_z) - 1;
	
	posf.xy = ex_coord * res;
	posi.x = int(floor(posf.x));
	posi.y = int(floor(posf.y));
	
	p.x = int(floor(4*(posf.x - posi.x))); // which part of the domain
	p.y = int(floor(4*(posf.y - posi.y)));
	p.z = int(floor(2*posf.z));
	
	f.x = 4*(posf.x - posi.x) - p.x; // fractional value in [0,1] of point in part
	f.y = 4*(posf.y - posi.y) - p.y;
	f.z = 2*posf.z - p.z;
	
	// convert to symmetric part
	bool flipx = false, flipy = false, flipxy = false;

	if (p.x >= 2)
	{
		flipx = true;
		p.x = 3 - p.x;
		f.x = 1 - f.x;
	}
	if (p.y >= 2)
	{
		flipy = true;
		p.y = 3 - p.y;
		f.y = 1 - f.y;
	}
	if (p.y > p.x)
	{
		flipxy = true;
		p.xy = p.yx;
		f.xy = f.yx;
	}

	int sind = (p.x + p.y + 3*p.z) * nfuncs;
	
	for (int i = 0; i < nfuncs; i++)
	{
		ivec3 o = indices_low[sind+i];
		
		int exp2oz = (1<<o.z);
		if (flipxy)
			o.xy = o.yx;
		
		if (flipx)
			o.x = (1<<o.z) - 1 - o.x;
		if (flipy)
			o.y = (1<<o.z) - 1 - o.y;

		o.xy += posi.xy*exp2oz;
		o.z = posi.z - o.z;

		vec4 cvec = coeffs_low[sind+i];
		float c = dot(cvec, vec4(1, f.x, f.y, f.z));
		
		sum += c * textureLod(un_tex, (vec2(o.xy)+.5)/(res*exp2oz), o.z).rgb;
	}

	return sum;
}

vec3 bilinear(vec2 pin, ivec2 size, int lev)
{
	vec2 p, t, f;
	
	for (int i = 0; i < 2; i++)
	{
		p[i] = pin[i] * size[i] - .5;
		f[i] = floor(p[i]);
		t[i] = p[i] - f[i];
	}
	
	float one = 1.0 / size[0];
	
	f += .5;
	f /= size[0];
	
	return  mix(
			mix(textureLod(un_tex, vec2(f.x,f.y), lev).rgb, textureLod(un_tex, vec2(f.x,f.y+one), lev).rgb, t.y),
			mix(textureLod(un_tex, vec2(f.x+one,f.y), lev).rgb, textureLod(un_tex, vec2(f.x+one,f.y+one), lev).rgb, t.y),
			t.x);
}

void main()
{		
	ivec2 ts = textureSize(un_tex, 0);
	int mipmax = int(log2(ts.x));
	
	
	
	
	
	/*vec2 dx = dFdx(ex_coord) * ts.x;
	vec2 dy = dFdy(ex_coord) * ts.y;
	float dd = max(length(dx),length(dy));
	ex_z = log2(dd);*/
	
	ex_z = textureQueryLod(un_tex, ex_coord).y;
	
	
	
	
	posi.z = int(ceil(ex_z)); // index of lowest resolution
	res = textureSize(un_tex, int(posi.z)).x; // lowest resolution
	
	if (ex_z >= mipmax)
	{
		gl_FragColor.rgb = texture(un_tex, ex_coord).rgb;
	}
	else if (ex_z <= 0)
	{
		//gl_FragColor.rgb = vec3(1,0,0); // DEBUG: bad. our method doesn't compare
		gl_FragColor.rgb = bilinear(ex_coord, ts, 0);
	}
	else if (ex_z <= 1)
	{
		gl_FragColor.rgb = sample_low();
	}
	else
	{
		gl_FragColor.rgb = sample_full();
	}
}
